% -------------------------------------------------------------------------
% Matlab Code for the Paper:
% "A novel diffeomorphic model for image registration and its algorithm"
% Journal of Mathematical Imaging and Vision 60 (2018): 1261-1283.
% Author: Daoping Zhang
%
% This implementation uses a matrix-based method for 2D image registration.
% The programming mechanism is modeled after the FAIR package.
% Some subfunctions, like interpolation, nodal2center and center, are adapted from the FAIR package:
% https://github.com/C4IR/FAIR.m
% -------------------------------------------------------------------------

clear 
close all
clc

%% Input images
T = imread('data/circle.jpg');
R = imread('data/c.jpg');

if ndims(T)==2
  dataT = double(flipud(T)');
else
  dataT = double(flipud(rgb2gray(T))');  
end

if ndims(R)==2
  dataR = double(flipud(R)');
else
  dataR = double(flipud(rgb2gray(R))');  
end

%% Input parameters
alpha = 100;
beta  = 10;

%% Image domain: it is fixed.
omega = [0 1 0 1];

%% Generate multilevel data from MinLevel to MaxLevel
MinLevel = 3;                   
MaxLevel = 7;                   
m = [2^MaxLevel,2^MaxLevel];    % the max resolution
MLdata = getMultileveldata(dataT,dataR,m,omega,MinLevel,MaxLevel);

%% run the image registration from minlevel to maxlevel
minlevel = 3;
maxlevel = 7;

%% Multilevel strategy to solve the proposed model
record_objectivevalue = zeros(maxlevel,1000);
record_normgrad       = zeros(maxlevel,1000);

tic
for Level = minlevel:maxlevel

   m  = MLdata{Level}.m;
   T  = MLdata{Level}.T;
   R  = MLdata{Level}.R;
   hd = prod((omega(2:2:end)-omega(1:2:end))./m); 
   P = nodal2center(m);

   [Dx,Dy] = getDiscreteGradient(m,omega);
   
   xc = getNodalGrid(omega,m);
   yRef = xc;
   
   if Level == minlevel
      y0 = yRef;
   else      
      % generate an initial guess for the next level
      U = zeros([m+1,2]);
      U(1:2:end,1:2:end,:)   = reshape(yc-xOld,[m/2+1,2]);
      U(2:2:end-1,1:2:end,:) = 0.5*(U(1:2:end-2,1:2:end,:)+U(3:2:end,1:2:end,:));
      U(:,2:2:end-1,:)       = 0.5*(U(:,1:2:end-2,:)+U(:,3:2:end,:));
     
      y0 = U(:)+xc;   
   end
    
   [yc,k,level_objc,level_ngrad]    = GaussNewton(T,R,m,omega,alpha,beta,y0,yRef,hd,P,Dx,Dy);
   record_objectivevalue(Level,1:k) = level_objc;
   record_normgrad(Level,1:k)       = level_ngrad;

   xOld = xc;
end
toc

plotresult(MLdata,Level,yc,yRef,omega,m,Dx,Dy)

